/**
 * @file MaxLloydQuantizer.h
 * @brief Declaration of the MaxLloydQuantizer class.
 *  Created on: 05.02.2009
 * @author Ruediger Knoerig
 */

#ifndef MAXLLOYDQUANTISIERER_H_
#define MAXLLOYDQUANTISIERER_H_

#include <cmath>
#include <itpp/itbase.h>
#include <itpp/stat/histogram.h>

using namespace std;

namespace itpp {

/**
 * @brief This class implements an optimal scalar Max-Lloyd-Quantizer.
 * @author Ruediger Knoerig
 */
template<class T>class MaxLloydQuantizer {
public:

	/**
	 * @brief Constructor resulting in an uninitialized and untrained quantizer.
	 */
	MaxLloydQuantizer() { initialized = trained = false; }

	/**
	 * @brief Constructor which initialize the quantizer.
	 * @param R Target rate (Number of quantizer levels = \f$2^R\f$).
	 */
	MaxLloydQuantizer(unsigned int R) { setTargetRate(R); }

	/**
	 * @brief Constructor which initialize and train the quantizer.
	 * @param R Target rate (Number of quantizer levels = \f$2^R\f$).
	 * @param trainingdata Data to train the quantizer.
	 * @param epsilon Terminate iteration if \f$|\vec{d}_{n+1}-\vec{d}_n|\leq \epsilon\f$
	 * @param max_iter Maximum number of iterations.
	 * @param hist_size Number of histogram bins
	 */
	MaxLloydQuantizer(unsigned int R,itpp::Vec<T> trainingdata,double epsilon=1e-3,unsigned int max_iter=1e3,unsigned hist_size=1000) { setTargetRate(R); train(trainingdata,epsilon,max_iter,hist_size); }

	/**
	 * @brief Destructor.
	 */
	virtual ~MaxLloydQuantizer() {}

	/**
	 * @brief training method.
	 * @param trainingdata Data to train the quantizer.
	 * @param epsilon Terminate iteration if \f$|\vec{d}_{n+1}-\vec{d}_n|\leq \epsilon\f$
	 * @param max_iter Maximum number of iterations.
	 * @return True if \f$|\vec{r}_{n+1}-\vec{r}_n|\leq \epsilon\f$ has been reached.
	 */
	bool train(const itpp::Vec<T> &trainingdata,double epsilon=1e-3,unsigned int max_iter=1e3);

	bool train2(const itpp::Vec<T> &trainingdata,double epsilon=1e-3,unsigned int max_iter=1e3);

	/**
	 * @brief Getter for the number of quantizer levels.
	 * @return Number of quantizer levels.
	 */
	unsigned int getNumberOfQuantizerLevels() const { return N; }

	/**
	 * @brief Getter for the target rate.
	 * @return Target rate of the quantizer.
	 */
	unsigned int getTargetRate() const { return Rate; }

	/**
	 * @brief Inspector for the trained state of the quantizer.
	 * @return True if the quantizer has been trained.
	 */
	bool isTrained() const { return trained; }

	/**
	 * @brief Inspector for the initialized state of the quantizer.
	 *
	 */
	bool isInitialized() const { return initialized; }

	/**
	 * @brief Setter for the target rate / number of quantization levels \f$N=2^R\f$.
	 * After calling this method, the quantizer is untrained again!
	 * @param R new target rate.
	 */
	void setTargetRate(unsigned int R)
	{
		Rate=R;
		N=1<<R;
		d.set_length(N+1);
		r.set_length(N);
		trained=false;
		initialized=true;
	}

	/**
	 * @brief Quantization method. Quantize an input sample.
	 * Input samples smaller than the smallest training value will be quantized to 0,
	 * input samples greater than the biggest training value will be quantized to N-1.
	 * @param Input sample.
	 * @return Quantized value of x.
	 */
	unsigned int quantize(T x);

	/**
	 * @brief Inverse quantization method (fast, without range checks).
	 * @param level quantizer level index.
	 * @return reconstruction value
	 */
	T inv_quant(unsigned int level);

protected:
	itpp::Vec<T> d; /**< Quantizer cell bounds */
	itpp::Vec<T> r; /**< reconstruction values */
	unsigned int Rate; /**< target rate for the quantizer */
	unsigned int N; /**< number of quantization levels */
	bool trained; /**< flag indicating if the quantizer has been trained */
	bool initialized; /**< flag indicating that the quantizer has been initialized */
};


template<class T> T MaxLloydQuantizer<T>::inv_quant(unsigned int level)
{
#ifdef DEBUG
	if(!initialized) cerr << "MaxLloydQuantizer<T>::inv_quant() called before initializing the quantizer!" << endl;
	if(!trained) cerr << "MaxLloydQuantizer<T>::inv_quant() called before training the quantizer!" << endl;
	if(level >= N) if(!trained) cerr << "MaxLloydQuantizer<T>::inv_quant(level): level out of range!" << endl;
#endif
	return r(level);
}


/*template<class T>  bool MaxLloydQuantizer<T>::train2(const itpp::Vec<T> &trainingdata,double epsilon,unsigned int max_iter,unsigned int hist_size)
{
#ifdef DEBUG
	if(!initialized) cerr << "MaxLloydQuantizer<T>::train() called before initializing the quantizer!" << endl;
#endif
	T minv,maxv;
	minv=maxv=trainingdata(0);
	for(const T* ptr=trainingdata._data();ptr < trainingdata._data()+trainingdata.length();ptr++)
	{
		register T val = *ptr;
		minv = ((val < minv) ? val : minv);
		maxv = ((val > maxv) ? val : maxv);
	}

	// berechne nun die Intervallgrenzen eines gleichverteilten Quantisierers
	T step=(maxv-minv)/N;
	T w=minv;
	for(unsigned int k=0;k<=N;k++,w+=step) d(k)=w;

	// nun erstelle die Wahrscheinlichkeitsdichteverteilung
	itpp::Histogram<double> hist(minv,maxv,hist_size);
	hist.update(trainingdata);
	Vec<T> p=hist.get_pdf();

	// nun erfolgt der eigentliche Max-Lloyd-Algorithmus
	step = (maxv-minv)/hist_size;
	unsigned int iteration=0;
	for(;iteration < max_iter;iteration++)
	{
		// berechne Rekonstruktionswerte neu
		for(unsigned int i=0;i<N;i++)
		{
			unsigned int llimit = std::floor((d(i)-minv)/step);
			unsigned int ulimit = std::floor((d(i+1)-minv)/step);
			T upperint=0;
			T lowerint=0;
			T x=d(i); //minv+llimit*step;
			for(unsigned int k=llimit;k<ulimit;k++,x+=step)
			{
				T val=p(k);
				lowerint += val;
				upperint += val*x;
			}
			r(i)=upperint/lowerint;
		}
		// berechne Intervallgrenzen neu
		double change=0.0;
		for(unsigned int i=1;i<N;i++)
		{
			T di_neu=(r(i-1)+r(i))/2;
			change += std::abs(double(d(i)-di_neu));
			d(i)=di_neu;
		}
		if(change/N <= epsilon) break;
	}

	trained=true;
	return (iteration==max_iter);
}*/

template<class T>  bool MaxLloydQuantizer<T>::train(const itpp::Vec<T> &trainingdata,double epsilon,unsigned int max_iter)
{
#ifdef DEBUG
	if(!initialized) cerr << "MaxLloydQuantizer<T>::train() called before initializing the quantizer!" << endl;
#endif
	if(trainingdata.length()==0) return false;
	T minv,maxv;
	minv=maxv=trainingdata(0);
	for(const T* ptr=trainingdata._data();ptr < trainingdata._data()+trainingdata.length();ptr++)
	{
		register T val = *ptr;
		minv = ((val < minv) ? val : minv);
		maxv = ((val > maxv) ? val : maxv);
	}

	// berechne nun die Intervallgrenzen eines gleichverteilten Quantisierers
	T step=(maxv-minv)/N;
	T w=minv;
	for(unsigned int k=0;k<=N;k++,w+=step) d(k)=w;
	w=minv+step/2;
	for(unsigned int k=0;k<N;k++,w+=step) r(k)=w;

	itpp::Vec<T> r_neu(N);
	itpp::Vec<unsigned int> r_anz(N);

	unsigned int iteration=0;
	for(;iteration < max_iter;iteration++)
	{
		r_neu.clear();
		r_anz.clear();
		// ordne jedem Trainingswert einen Rekonstruktionswert zu
		for(int k=0;k<trainingdata.length();k++)
		{
			T val=trainingdata(k);
			T min_dist=std::abs(r(0)-val);
			unsigned min_index=0;
			for(unsigned int l=1;l<N;l++)
			{
				T dist=std::abs(T(r(l)-val));
				if(dist < min_dist)
				{
					min_dist=dist;
					min_index=l;
				}
			}
			r_neu(min_index)+=val;
			r_anz(min_index)++;
		}
		double change=0.0;
		for(unsigned int k=0;k<N;k++)
		{
			double r_neuw = (r_anz(k) != 0 ? r_neu(k)/r_anz(k) : r(k));
			double tmp=r_neuw-r(k);
			change+=tmp*tmp;
			r(k)=r_neuw;
		}

		//cout << "Iteration " << iteration << " var: " << change << endl <<  " r=" << r << endl;

		if(change/N <= epsilon) break;
	}
	// berechne Intervallgrenzen neu

	for(unsigned int i=1;i<N;i++)
	{
		T di_neu=(r(i-1)+r(i))/2;

		d(i)=di_neu;
	}
	trained=true;
	return (iteration!=max_iter);
}

template<class T>  bool MaxLloydQuantizer<T>::train2(const itpp::Vec<T> &trainingdata,double epsilon,unsigned int max_iter)
{
#ifdef DEBUG
	if(!initialized) cerr << "MaxLloydQuantizer<T>::train() called before initializing the quantizer!" << endl;
#endif
	if(trainingdata.length()==0) return false;
	T minv,maxv;
	minv=maxv=trainingdata(0);
	for(const T* ptr=trainingdata._data();ptr < trainingdata._data()+trainingdata.length();ptr++)
	{
		register T val = *ptr;
		minv = ((val < minv) ? val : minv);
		maxv = ((val > maxv) ? val : maxv);
	}

	// berechne nun die Intervallgrenzen eines gleichverteilten Quantisierers
	T step=(maxv-minv)/N;
	T w=minv;
	for(unsigned int k=0;k<N;k++,w+=step) d(k)=w;
	// zur Sicherheit
	d(N)=maxv;
	w=minv+step/2;
	for(unsigned int k=0;k<N;k++,w+=step) r(k)=w;

	itpp::Vec<T> r_neu(N);
	itpp::Vec<unsigned int> r_anz(N);

	unsigned int iteration=0;
	for(;iteration < max_iter;iteration++)
	{
		r_neu.clear();
		r_anz.clear();
		// ordne jedem Trainingswert einen Rekonstruktionswert zu
		for(int k=0;k<trainingdata.length();k++)
		{
			T val=trainingdata(k);
			for(unsigned int l=1;l<=N;l++)
			{
				if(val <= d(l))
				{
					r_neu(l-1)+=val;
					++r_anz(l-1);
					break;
				}
			}
		}
		double change=0.0;
		for(unsigned int k=0;k<N;k++)
		{
			r(k) = (r_anz(k) != 0 ? r_neu(k)/r_anz(k) : 0.5*(d(k)+d(k+1)));
		}

		// berechne Intervallgrenzen neu
		for (unsigned int i = 1; i < N; i++) {
			T di_neu = (r(i - 1) + r(i)) / 2;
			change+=std::abs((double) di_neu-d(i));
			d(i) = di_neu;
		}
		cout << "Iteration " << iteration << " var: " << change << endl <<  " r=" << r << endl;

		if(change/N <= epsilon) break;
	}

	trained=true;
	return (iteration!=max_iter);
}


template<class T> unsigned int MaxLloydQuantizer<T>::quantize(T x)
{
#ifdef DEBUG
	if(!initialized) cerr << "MaxLloydQuantizer<T>::quant() called before initializing the quantizer!" << endl;
	if(!trained) cerr << "MaxLloydQuantizer<T>::quant() called before training the quantizer!" << endl;
#endif
	unsigned int k=0;
	while(++k != N) if(x < d(k)) break;
	return k-1;
}

}
#endif /* MAXLLOYDQUANTISIERER_H_ */

/** \example main.cpp
 * Example code for using this class.
 */

/*! \mainpage Hauptseite
 *
 * \section intro_sec Introduction
 * The MaxLloydQuantizer class implements an optimal scalar quantizer based on the Max-Lloyd-Algorithm.
 */
